import pandas as pd
import numpy as np
from sklearn.metrics import make_scorer, confusion_matrix, classification_report, f1_score, fbeta_score, roc_auc_score, auc as sk_auc, roc_curve, precision_score, recall_score
import matplotlib.pyplot as plt

def add_actual_rolling(df, win):
    df["y_actual"] = (df["y_true"].diff()==-1).astype(int)
    df["preds_rolling"] = df["preds"].rolling(window=win).mean().fillna(0)
    return df


def evaluate_f1(df, threshold):
    df["preds_binary"] = (df["preds_rolling"]>threshold).astype(int)
    res = f1_score(df["y_actual"], df["preds_binary"])
    return res


def evaluate_fbeta(df, threshold, beta):
    df["preds_binary"] = (df["preds_rolling"]>threshold).astype(int)
    res = fbeta_score(df["y_actual"], df["preds_binary"], beta=beta)
    return res


def evaluate_cm(df, threshold):
    df["preds_binary"] = (df["preds_rolling"]>threshold).astype(int)
    cm = confusion_matrix(df["y_actual"], df["preds_binary"])
    return cm


def evaluate_classification_report(df, threshold):
    res = classification_report(df["y_actual"], df["preds_binary"], output_dict=True)
    return res["1"]

def f1_actual(y_true, y_pred, win):
    df = pd.DataFrame()
    df["y_true"] = y_true
    df["preds"] = y_pred
    df = add_actual_rolling(df, win*2)
    scores = {t: evaluate_f1(df, t) for t in np.linspace(0,1, 101)}
    best_t = max(scores, key=lambda x: scores[x])
    return max(list(scores.values()))

def find_best_f1(y_true, y_pred, win):
    df = pd.DataFrame()
    df["y_true"] = y_true
    df["preds"] = y_pred
    df = add_actual_rolling(df, win*2)
    scores = {t: evaluate_f1(df, t) for t in np.linspace(0,1, 101)}
    best_t = max(scores, key=lambda x: scores[x])
    cm = evaluate_cm(df, best_t)
    res = {m:v for (m, v) in zip(["tn","fp", "fn", "tp"], cm.ravel())}
    res["threshold"] = best_t
    res.update(**evaluate_classification_report(df, best_t))
    return res

def find_best_fbeta(y_true, y_pred, win, beta=4):
    df = pd.DataFrame()
    df["y_true"] = y_true
    df["preds"] = y_pred
    df = add_actual_rolling(df, win*2)
    scores = {t: evaluate_fbeta(df, t, beta) for t in np.linspace(0,1, 101)}
    best_t = max(scores, key=lambda x: scores[x])
    cm = evaluate_cm(df, best_t)
    res = {m:v for (m, v) in zip(["tn","fp", "fn", "tp"], cm.ravel())}
    res["threshold"] = best_t
    res.update(**evaluate_classification_report(df, best_t))
    return res

def save_roc(y_true, y_pred, model_name):
    fpr, tpr, _ = roc_curve(y_true, y_pred)
    roc_auc = roc_auc_score(y_true, y_pred)
    plt.figure()
    lw = 2
    plt.plot(fpr, tpr, color='darkorange',
             lw=lw, label='ROC curve (area = %0.2f)' % roc_auc)
    plt.plot([0, 1], [0, 1], color='navy', lw=lw, linestyle='--')
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    #plt.title('Receiver operating characteristic ')
    plt.legend(loc="lower right")
    plt.savefig(model_name)
    plt.clf()
    return roc_auc

def get_result_df(results):
    try:
        df = pd.DataFrame.from_dict(results)
    except:
        df = pd.DataFrame()
        df["num_samples"] = [5770]*5
        df["num_failures"] = [4]*5
        df["threshold"] = [0.17, 0.66, 0.96, 0.64, 0.3]
        df["failures_detected"] = [1,1,1,1,3]
        df["false_alarms_generated"] = [3,25,6,22,95]
        df["precision"] = df["failures_detected"]/(df["false_alarms_generated"]+df["failures_detected"])
        df["recall"] = df["failures_detected"]/df["num_failures"]
        df["f1-score"] = 2*((df["precision"].multiply(df["recall"]))/(df["precision"]+df["recall"]))
        df.index = ["rf", "mlp", "lstm", "fcn", "tcn"]
    return df